#ifndef RLCSA_H
#define RLCSA_H

#include <fstream>
#include <vector>

#include "bits/vectors.h"
#include "sasamples.h"
#include "misc/parameters.h"


namespace CSA
{


const std::string SA_EXTENSION = ".sa";
const std::string PSI_EXTENSION = ".psi";
const std::string ARRAY_EXTENSION = ".rlcsa.array";
const std::string SA_SAMPLES_EXTENSION = ".rlcsa.sa_samples";
const std::string PARAMETERS_EXTENSION = ".rlcsa.parameters";


const parameter_type RLCSA_BLOCK_SIZE  = parameter_type("RLCSA_BLOCK_SIZE", 32);
const parameter_type SAMPLE_RATE       = parameter_type("SAMPLE_RATE", 512);
const parameter_type SUPPORT_LOCATE    = parameter_type("SUPPORT_LOCATE", 1);
const parameter_type SUPPORT_DISPLAY   = parameter_type("SUPPORT_DISPLAY", 1);


class RLCSA
{
  public:

//--------------------------------------------------------------------------
//  CONSTRUCTION
//--------------------------------------------------------------------------

    const static usint ENDPOINT_BLOCK_SIZE = 16;

    RLCSA(const std::string& base_name, bool print = false);

    RLCSA(RLCSA* original);

    /*
      If multiple_sequences is true, each 0 is treated as a end of sequence marker.
      There must be nonzero characters between the 0s. data[bytes - 1] must also be 0.
      FIXME Crashes if bytes >= 2 GB.
    */ 
    RLCSA(uchar* data, usint bytes, usint block_size, usint sa_sample_rate, bool multiple_sequences, bool delete_data);

    // Destroys contents of index and increment.
    // threads has no effect unless MULTITHREAD_SUPPORT is defined
    RLCSA(RLCSA& index, RLCSA& increment, usint* positions, usint block_size, usint threads = 1);
    ~RLCSA();

    void writeTo(const std::string& base_name);

    bool isOk();

//--------------------------------------------------------------------------
//  OPERATIONS
//--------------------------------------------------------------------------

    // Returns the closed interval of suffix array containing the matches.
    pair_type count(const std::string& pattern);

    // Returns the number of matches with k mismatching characters.
    // The actual match ranges are inserted into the vector given.
    // Time complexity scales as sigma^k.
    // FIXME Should probably parallelize.
    // FIXME Should probably recurse only on mismatches.
    usint kMismatches(const std::string& pattern, usint k, std::vector<pair_type>& ranges);

    // Used when merging CSAs.
    void reportPositions(uchar* data, usint length, usint* positions);

    // Generates text using a Markov model.
    usint generateText(std::string context, usint context_length, uchar* data, usint max_length);

    // Returns SA[range]. User must free the buffer. Latter version uses buffer provided by the user.
    usint* locate(pair_type range);
    usint* locate(pair_type range, usint* data);

    // Returns SA[index].
    usint locate(usint index);

    // Returns T^{sequence}[range]. User must free the buffer.
    // Third version uses buffer provided by the user.
    uchar* display(usint sequence);
    uchar* display(usint sequence, pair_type range);
    uchar* display(usint sequence, pair_type range, uchar* data);

    // Displays the intersection of T[position - context, position + len + context - 1]
    // and T^{getSequenceForPosition(position)}.
    // This is intended for displaying an occurrence of a pattern of length 'len'
    // with 'context' extra characters on both sides.
    // The actual length of the returned string is written into result_length.
    uchar* display(usint position, usint len, usint context, usint& result_length);

    // Returns the actual length of the prefix. User must provide the buffer.
    usint displayPrefix(usint sequence, usint len, uchar* data);

    // Get the range of SA values for the sequence identified by
    // a sequence number or a SA value.
    pair_type getSequenceRange(usint number);
    pair_type getSequenceRangeForPosition(usint value);

    // Get the sequence number for given SA value(s).
    // The returned array is the same as the parameter.
    usint getSequenceForPosition(usint value);
    usint* getSequenceForPosition(usint* value, usint length);

    // Writes the Psi array into given file. End of sequence markers are not written.
    // void decompressInto(std::ofstream& psi_file);

    // Returns the BWT of the collection including end of sequence markers.
    uchar* readBWT();
    uchar* readBWT(pair_type range);

    // Return value includes the implicit end of sequence markers. To get suffix array indexes,
    // subtract getNumberOfSequences() from the value.
    usint psi(usint index);
    pair_type psi(usint index, usint max_length); // This version returns a run.

//--------------------------------------------------------------------------
//  REPORTING
//--------------------------------------------------------------------------

    inline bool supportsLocate()       { return this->support_locate; }
    inline bool supportsDisplay()      { return this->support_display; }
    inline usint getSize()              { return this->data_size; }
    inline usint getNumberOfSequences() { return this->number_of_sequences; }
    inline usint getBlockSize() { return this->array[this->text_chars[0]]->getBlockSize(); }

    // Returns the size of the data structure.
    usint reportSize(bool print = false);

    void printInfo();

//--------------------------------------------------------------------------
//  INTERNAL STUFF
//--------------------------------------------------------------------------

  private:
    bool ok;

    RLEVector* array[CHARS];
    SASamples* sa_samples;

    pair_type index_ranges[CHARS];
    usint data_size;

    usint text_chars[CHARS];  // which characters are present in the text
    usint chars;

    usint index_pointers[CHARS]; // which of the above is at i * index_rate
    usint index_rate;

    bool support_locate, support_display;
    usint sample_rate;

    usint number_of_sequences;
    DeltaVector* end_points;

    inline pair_type backwardSearchStep(pair_type range, usint c)
    {
      RLEVector* vect = this->array[c];
      if(vect == 0) { return EMPTY_PAIR; }

      usint start = this->index_ranges[c].first + this->number_of_sequences - 1;
      range.first = start + vect->rank(range.first, true);
      range.second = start + vect->rank(range.second);

      return range;
    }

    usint kMismatches(const std::string& pattern, usint len, usint k, std::vector<pair_type>& ranges, pair_type search_range);

    void locateUnsafe(pair_type range, usint* data);
    bool processRun(pair_type run, usint* data, usint* offsets, bool* finished);
    void displayUnsafe(pair_type range, uchar* data);

    inline usint getCharacter(usint pos)
    {
      usint* curr = &(this->text_chars[this->index_pointers[pos / this->index_rate]]);
      while(pos > this->index_ranges[*curr].second) { curr++; }
      return *curr;
    }

    void mergeEndPoints(RLCSA& index, RLCSA& increment);
    void mergeSamples(RLCSA& index, RLCSA& increment, usint* positions);

    void buildCharIndexes(usint* distribution);
};


// Returns the total number of characters.
usint buildRanges(usint* distribution, pair_type* index_ranges);


} // namespace CSA


#endif // RLCSA_H
